import polars as pl

from .abstract_executor import AbstractPythonExecutor


class PolarsLocalExecutor(AbstractPythonExecutor[pl.DataFrame]):
    library_name = "polars"
    handled_types = ["PolarsDbt"]

    def prepare_bulk_write(self, df: pl.DataFrame, table: str, schema: str) -> str:
        df.head(1).write_database(
            table_name=f"{schema}.{table}",
            connection=self.conn_string,
            if_table_exists="replace",
            engine="adbc",
        )
        return df.write_csv(separator="\t", null_value="\\N", include_header=False)



from dataclasses import dataclass


@dataclass
class ExecutionResult:
    """Result from executing a dbt Python model."""

    rows_affected: int
    table_name: str
    schema: str
    table: str

    def __str__(self) -> str:
        return f"SELECT {self.rows_affected:,}"



from .ast_utils import extract_python_docstring, get_library_from_typehint
from .general import find_funcs_in_stack, logs, release_plugin_lock
from .profile_parsing import find_profile, find_target

__all__ = [
    "extract_python_docstring",
    "get_library_from_typehint",
    "find_funcs_in_stack",
    "release_plugin_lock",
    "logs",
    "find_profile",
    "find_target",
]



import geopandas as gpd
import pandas as pd

from .base import DbtConfig, DbtThis, SessionObject
from .dbt_objects import DbtObject, PolarsDataFrame

PandasDbt = DbtObject[pd.DataFrame]
"""Type hint for pandas-based dbt models."""

GeoPandasDbt = DbtObject[gpd.GeoDataFrame]
"""Type hint for GeoPandas-based dbt models."""

PolarsDbt = DbtObject[PolarsDataFrame]
"""Type hint for Polars-based dbt models (DataFrame or LazyFrame)."""


__all__ = [
    "PolarsDataFrame",
    "DbtThis",
    "DbtConfig",
    "DbtObject",
    "PandasDbt",
    "PolarsDbt",
    "GeoPandasDbt",
    "SessionObject",
]



import pandas as pd
from sqlalchemy import create_engine

from .abstract_executor import AbstractPythonExecutor


class PandasPythonExecutor(AbstractPythonExecutor[pd.DataFrame]):
    library_name = "pandas"
    handled_types = ["PandasDbt"]

    def prepare_bulk_write(self, df: pd.DataFrame, table: str, schema: str) -> str:
        engine = create_engine(self.conn_string)
        df.head(1).to_sql(
            name=table, con=engine, schema=schema, if_exists="replace", index=False
        )
        return df.to_csv(sep="\t", na_rep="\\N", index=False, header=False)



from .abstract_executor import AbstractPythonExecutor
from .geo_pandas_executor import GeoPandasLocalExecutor
from .pandas_executor import PandasPythonExecutor
from .polars_local_executor import PolarsLocalExecutor
from .result import ExecutionResult

__all__ = [
    "PandasPythonExecutor",
    "PolarsLocalExecutor",
    "AbstractPythonExecutor",
    "ExecutionResult",
    "GeoPandasLocalExecutor",
]



"""Main CLI module for dbt-depp using cyclopts."""

from typing import Annotated, Optional

import cyclopts

app = cyclopts.App(
    name="dbt-depp",
    help="DBT Enexis Python Postgres adapter - Run python scripts from any dbt project.",
)


@app.command
def run(
    model: Annotated[Optional[str], cyclopts.Parameter(help="Model to run")] = None,
    profile: Annotated[str, cyclopts.Parameter(help="Profile to use")] = "default",
    verbose: Annotated[bool, cyclopts.Parameter(help="Enable verbose output")] = False,
) -> None:
    """Run a dbt model with the depp adapter."""
    if verbose:
        print(f"Running model: {model} with profile: {profile}")
    else:
        print(f"Running {model}")


@app.command
def compile(
    model: Annotated[Optional[str], cyclopts.Parameter(help="Model to compile")] = None,
    profile: Annotated[str, cyclopts.Parameter(help="Profile to use")] = "default",
) -> None:
    """Compile a dbt model."""
    print(f"Compiling model: {model} with profile: {profile}")


@app.command
def test(
    model: Annotated[Optional[str], cyclopts.Parameter(help="Model to test")] = None,
    profile: Annotated[str, cyclopts.Parameter(help="Profile to use")] = "default",
) -> None:
    """Run tests for a dbt model."""
    print(f"Testing model: {model} with profile: {profile}")


def main() -> None:
    """Entry point for the CLI."""
    app()


if __name__ == "__main__":
    main()



from functools import partial
from typing import TYPE_CHECKING, Callable

from ..utils.general import find_funcs_in_stack

if TYPE_CHECKING:
    from ..adapter import DeppAdapter

ADAPTER_NAME = "depp"


class AdapterTypeDescriptor:
    # TODO: can we use this in a more general way for all type things like connections?
    type_str: str = ADAPTER_NAME

    def __get__(
        self, obj: "DeppAdapter | None", objtype: type["DeppAdapter"] | None = None
    ) -> Callable[[], str] | partial[str]:
        def _type(instance: "DeppAdapter | None" = None) -> str:
            if instance is None:
                return ADAPTER_NAME
            if find_funcs_in_stack({"render", "db_materialization"}):
                return instance.db_adapter.type()
            return ADAPTER_NAME

        return partial(_type, obj) if obj else _type



from dataclasses import dataclass
from typing import Any, Dict, Tuple

from dbt.adapters.contracts.connection import Credentials

from .adapter_type import ADAPTER_NAME


@dataclass
class DeppCredentials(Credentials):
    db_profile: str = ""

    def _connection_keys(self) -> Tuple[str, ...]:
        return ("db_profile",)

    @property
    def type(self) -> str:
        return ADAPTER_NAME

    @property
    def unique_field(self) -> str:
        return self.db_profile

    @classmethod
    def translate_aliases(
        cls, kwargs: Dict[str, Any], recurse: bool = False
    ) -> Dict[str, Any]:
        return kwargs



import inspect
import time
from contextlib import contextmanager
from functools import wraps
from types import FrameType
from typing import TYPE_CHECKING, Callable, Concatenate, Iterator, ParamSpec, TypeVar

from dbt.adapters.contracts.connection import AdapterResponse
from dbt.adapters.events.types import CodeExecution, CodeExecutionStatus
from dbt.adapters.factory import FACTORY
from dbt_common.events.functions import fire_event

if TYPE_CHECKING:
    from ..adapter import DeppAdapter

T = TypeVar("T", bound="DeppAdapter")
P = ParamSpec("P")

funcT = Callable[Concatenate[T, P], AdapterResponse]


def logs(func: funcT[T, P]) -> funcT[T, P]:
    """Decorator for python executor methods to log"""

    @wraps(func)
    def logs(self: T, *args: P.args, **kwargs: P.kwargs) -> AdapterResponse:
        connection_name = self.connections.get_thread_connection().name
        compiled_code = args[1]
        fire_event(CodeExecution(conn_name=connection_name, code_content=compiled_code))

        start_time = time.time()
        response = func(self, *args, **kwargs)
        elapsed = round((time.time() - start_time), 2)

        fire_event(CodeExecutionStatus(status=response.__str__(), elapsed=elapsed))
        return response

    return logs  # type: ignore[return-value]


def find_funcs_in_stack(funcs: set[str]) -> bool:
    frame: FrameType | None = inspect.currentframe()
    while frame:
        if frame.f_code.co_name in funcs:
            return True
        frame = frame.f_back
    return False


@contextmanager
def release_plugin_lock() -> Iterator[None]:
    FACTORY.lock.release()
    try:
        yield
    finally:
        FACTORY.lock.acquire()



from .adapter_type import AdapterTypeDescriptor
from .connections import DeppCredentials
from .credential_wrapper import DeppCredentialsWrapper
from .model_config import ModelConfig
from .profile_loader import DbInfo, RelationDescriptor

__all__ = [
    "AdapterTypeDescriptor",
    "DeppCredentials",
    "DeppCredentialsWrapper",
    "RelationDescriptor",
    "DbInfo",
    "ModelConfig",
]



from dbt.adapters.base.plugin import AdapterPlugin

from ...include import depp  # type: ignore
from .adapter import DeppAdapter
from .config import DeppCredentials


def __getattr__(name: str) -> AdapterPlugin:
    return AdapterPlugin(
        adapter=DeppAdapter,  # type: ignore
        credentials=DeppCredentials,
        include_path=depp.PACKAGE_PATH,
    )



import io
import re

import connectorx as cx
import geopandas as gpd
import numpy as np
from geoalchemy2 import Geometry
from geopandas.array import GeometryDtype
from sqlalchemy import create_engine

from .abstract_executor import AbstractPythonExecutor, SourceInfo


class GeoPandasLocalExecutor(AbstractPythonExecutor[gpd.GeoDataFrame]):
    library_name = "geopandas"
    handled_types = ["GeoPandasDbt"]

    @staticmethod
    def detect_geometry_column(value: str) -> bool:
        """Check if a string value looks like WKT geometry."""
        if not value or not value.strip():
            return False
        return bool(
            re.match(
                r"^(POINT|LINESTRING|POLYGON|MULTI|GEOMETRY)", value.strip().upper()
            )
        )

    def prepare_bulk_write(self, df: gpd.GeoDataFrame, table: str, schema: str) -> str:
        engine = create_engine(self.conn_string)
        dtype_mapping = {
            col: Geometry("GEOMETRY", srid=4326)
            for col, dtype in df.dtypes.items()
            if dtype == "geometry"
        }
        df.head(1).to_postgis(
            name=table,
            con=engine,
            schema=schema,
            if_exists="replace",
            index=False,
            dtype=dtype_mapping,
        )

        df_copy = df.copy()
        geom_cols = [
            col
            for col in df_copy.columns
            if isinstance(df_copy[col].dtype, GeometryDtype)
        ]
        for col in geom_cols:
            df_copy[col] = df_copy[col].to_wkt().fillna("\\N")

        output = io.StringIO()
        np_array = df_copy.to_numpy()
        np.savetxt(output, np_array, delimiter="\t", fmt="%s", newline="\n")
        return output.getvalue()

    def _get_geometry_columns(
        self, schema: str, table: str
    ) -> tuple[dict[str, str], int]:
        query = f"""
            SELECT f_geometry_column as col_name, type as geom_type, srid
            FROM geometry_columns 
            WHERE f_table_schema = '{schema}' AND f_table_name = '{table}'
        """
        geom_df = cx.read_sql(self.conn_string, query)  # type: ignore
        srid = int(geom_df["srid"].iloc[0])
        return dict(zip(geom_df["col_name"], geom_df["geom_type"])), srid

    def _get_all_columns(self, source: SourceInfo) -> list[str]:
        # TODO: Find out if this can cause deadlocks on information_schema
        print(source.full_name)
        cols_query = f"""
            SELECT column_name 
            FROM information_schema.columns 
            WHERE (table_schema || '.' || table_name = '{source.full_name}'
                   OR table_name = '{source.full_name}')
            ORDER BY ordinal_position
        """
        return cx.read_sql(self.conn_string, cols_query)["column_name"].tolist()  # type: ignore

    def read_df(self, table_name: str) -> gpd.GeoDataFrame:
        """Read PostGIS table."""
        # TODO: add support for providing geometry to reduce queries
        source = self.get_source_info(table_name)
        geom_cols, srid = self._get_geometry_columns(source.schema, source.table)
        all_cols = self._get_all_columns(source)
        regular_cols = [c for c in all_cols if c not in geom_cols]
        select_parts = regular_cols + [
            f"ST_AsBinary({c}) as {c}_wkb" for c in geom_cols
        ]
        query = f"SELECT {', '.join(select_parts)} FROM {table_name}"
        df = cx.read_sql(self.conn_string, query, protocol="binary")  # type: ignore
        wkb_cols = [f"{col}_wkb" for col in geom_cols]
        for col in geom_cols:
            df[col] = gpd.GeoSeries.from_wkb(df[f"{col}_wkb"])
        df = df.drop(columns=wkb_cols)
        crs = f"EPSG:{srid}" if srid else None
        return gpd.GeoDataFrame(df, geometry=list(geom_cols)[0], crs=crs)



from typing import Any

from dbt.config.project import load_raw_project
from dbt.config.renderer import ProfileRenderer


def find_profile(
    override: str | None, root: str, render: ProfileRenderer
) -> str | None:
    if override is not None:
        return override

    raw_profile = load_raw_project(root).get("profile")
    result = render.render_value(raw_profile)
    return result if isinstance(result, str) else None


def find_target(
    override: str | None, profile: dict[str, Any], render: ProfileRenderer
) -> str:
    if override is not None:
        return override
    if "target" in profile:
        result = render.render_value(profile["target"])
        return result if isinstance(result, str) else "default"
    return "default"



from argparse import Namespace
from dataclasses import dataclass
from functools import lru_cache
from typing import TYPE_CHECKING, Any

from dbt.adapters.factory import FACTORY
from dbt.adapters.protocol import RelationProtocol
from dbt.config.profile import Profile, read_profile
from dbt.config.renderer import ProfileRenderer
from dbt.flags import get_flags

from ..utils import find_profile, find_target

if TYPE_CHECKING:
    from dbt.config.runtime import RuntimeConfig

    from ..adapter import DeppAdapter


@dataclass
class DbInfo:
    profile: Profile
    override_properties: dict[str, Any]
    relation: RelationProtocol | None = None

    @classmethod
    def load_profile_info(cls) -> "DbInfo":
        """Load database profile from depp adapter configuration"""
        # TODO: some of this code feels like it could use an upgrade
        flags: Namespace = get_flags()  # type: ignore
        renderer = ProfileRenderer(getattr(flags, "VARS", {}))

        name = find_profile(flags.PROFILE, flags.PROJECT_DIR, renderer)
        if name is None:
            raise ValueError("Profile name not found")
        profile = read_profile(flags.PROFILES_DIR)[name]
        target_name = find_target(flags.TARGET, profile, renderer)
        _, depp_dict = Profile.render_profile(profile, name, target_name, renderer)

        if not (db_target := depp_dict.get("db_profile")):
            raise ValueError("depp credentials must have a `db_profile` property set")

        try:
            db_profile = Profile.from_raw_profile_info(
                profile, name, renderer, db_target
            )
        except RecursionError as e:
            raise AttributeError("Cannot nest depp profiles within each other") from e

        threads = getattr(
            flags, "THREADS", depp_dict.get("threads") or db_profile.threads
        )
        override_properties = dict(threads=threads)
        return cls(db_profile, override_properties)

    @classmethod
    @lru_cache(maxsize=1)
    def get_cached_with_relation(cls) -> "DbInfo":
        """Get cached DbInfo instance with relation populated"""
        db_info = cls.load_profile_info()
        relation = FACTORY.get_relation_class_by_name(db_info.profile.credentials.type)
        return cls(db_info.profile, db_info.override_properties, relation)

    def apply_overrides(self, config: "RuntimeConfig") -> None:
        """Apply override properties to the given config object"""
        if self.override_properties:
            for key, value in self.override_properties.items():
                if value is not None:
                    setattr(config, key, value)


class RelationDescriptor:
    """Descriptor that lazily loads and caches the Relation class"""

    def __init__(self) -> None:
        self._relation: RelationProtocol | None = None

    def __get__(self, instance: "DeppAdapter | None", owner: "type[DeppAdapter] | None") -> RelationProtocol | None:
        if self._relation is None:
            self._relation = DbInfo.get_cached_with_relation().relation
        return self._relation



from typing import TYPE_CHECKING

import pandas as pd

if TYPE_CHECKING:
    from ...src.dbt.adapters.depp.typing import PandasDbt, SessionObject


def model(dbt: "PandasDbt", session: "SessionObject"):
    """
    Creates a simple customer table with sample data.

    This model generates basic customer information including IDs, names,
    regions, and total spending amounts for testing purposes.
    """
    # Create a simple dataframe with customer data
    data = {
        "customer_id": [1, 2, 3, 4, 5],
        "customer_name": ["Alice", "Bob", "Charlie", "Diana", "Eve"],
        "region": ["North", "South", "East", "West", "North"],
        "total_spent": [1500.00, 2300.00, 1800.00, 3200.00, 900.00],
    }

    df = pd.DataFrame(data)
    return df



"""Example using type narrowing helper for better IDE support."""

from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from ...src.dbt.adapters.depp.typing import GeoPandasDbt, SessionObject


def model(dbt: "GeoPandasDbt", session: "SessionObject"):
    products_df = dbt.source("raw_data", "result_table")
    return products_df



import asyncio
import io
from abc import ABC, abstractmethod
from dataclasses import dataclass
from typing import Any, Generic, TypeVar

import asyncpg  # type: ignore[import-untyped]
import connectorx as cx
from asyncpg import Connection
from dbt.adapters.contracts.connection import Credentials
from dbt.adapters.postgres.connections import PostgresCredentials

from .result import ExecutionResult

DataFrameType = TypeVar("DataFrameType")


@dataclass
class SourceInfo:
    full_name: str
    schema: str
    table: str


class AbstractPythonExecutor(ABC, Generic[DataFrameType]):
    """
    Base executor for running Python models in dbt with different DataFrame libraries.
    Subclasses must set `library_name` and implement `prepare_bulk_write`.
    Auto-registers implementations to a central registry.
    """

    registry: dict[str, type["AbstractPythonExecutor[Any]"]] = {}
    type_mapping: dict[str, str] = {}
    library_name: str | None = None
    handled_types: list[str] = []

    def __init_subclass__(cls, **kwargs: Any):
        """Automatically registers the executor to a registry"""
        super().__init_subclass__(**kwargs)
        if cls.library_name is not None:
            AbstractPythonExecutor.registry[cls.library_name] = cls
            # Build the type mapping from handled_types
            for type_hint in getattr(cls, "handled_types", []):
                AbstractPythonExecutor.type_mapping[type_hint] = cls.library_name

    @classmethod
    def get_library_for_type(cls, type_hint: str) -> str | None:
        """Get the library name for a given type hint."""
        return cls.type_mapping.get(type_hint)

    def __init__(self, parsed: dict[str, Any], db: Credentials, lib: str = "polars"):
        if not isinstance(db, PostgresCredentials):
            raise ValueError("Currently just postgres is supported")
        if lib not in self.registry:
            raise ValueError(f"Only {', '.join(self.registry)} supported")

        self.parsed_model = parsed
        self.library = lib
        self.conn_string = self.get_connection_string(db)

    @staticmethod
    def get_source_info(table_name: str) -> SourceInfo:
        clean_name = table_name.replace('"', "")
        _, schema, table = clean_name.split(".")
        return SourceInfo(f"{schema}.{table}", schema, table)

    def read_df(self, table_name: str) -> DataFrameType:
        # TODO: Support filtering
        """Reads all data from the table using connectorx"""
        source = self.get_source_info(table_name)
        query = f'SELECT * FROM "{source.schema}"."{source.table}"'

        # connectorx is untyped
        result = cx.read_sql(
            self.conn_string, query, return_type=self.library, protocol="binary"
        )
        return result  # type: ignore[return-value]

    def write_df(self, table_name: str, dataframe: DataFrameType) -> ExecutionResult:
        """Write DataFrame to database table using async bulk copy."""
        return asyncio.run(self.async_write_df(table_name, dataframe))

    @abstractmethod
    def prepare_bulk_write(self, df: DataFrameType, table: str, schema: str) -> str:
        """
        Convert DataFrame to TSV string for PostgreSQL COPY command.

        Implementations should:
        1. Ensure table exists (e.g., write single row first)
        2. Convert DataFrame to TSV format with \\N for nulls

        Args:
            df: DataFrame to prepare
            table: Target table name
            schema: Target schema name

        Returns:
            TSV-formatted string ready for COPY
        """
        raise NotImplementedError

    async def async_write_df(
        self, table_name: str, df: DataFrameType
    ) -> ExecutionResult:
        """Asynchronously bulk write DataFrame using PostgreSQL COPY."""

        parts = table_name.replace('"', "").split(".")
        schema, table = parts[-2], parts[-1]

        conn: Connection = await asyncpg.connect(dsn=self.conn_string)

        csv_buffer = io.BytesIO()
        csv_string = self.prepare_bulk_write(df, table, schema)
        csv_buffer.write(csv_string.encode("utf-8"))
        csv_buffer.seek(0)

        rows_copied: str = await conn.copy_to_table(
            table_name=table,
            schema_name=schema,
            source=csv_buffer,
            format="csv",
            delimiter="\t",
            null="\\N",
        )

        await conn.close()

        # Extract row count from COPY result
        rows_effected = int(rows_copied.split(sep=" ")[-1])
        return ExecutionResult(
            rows_affected=rows_effected,
            table_name=table_name,
            schema=schema,
            table=table,
        )

    def submit(self, compiled_code: str) -> ExecutionResult:
        """Execute compiled dbt Python model code."""
        local_vars: dict[str, Any] = {}
        exec(compiled_code, local_vars)
        if "main" not in local_vars:
            raise RuntimeError("No main function found in compiled code")
        result = local_vars["main"](self.read_df, self.write_df)
        return ExecutionResult(**result) if isinstance(result, dict) else result

    @staticmethod
    def get_connection_string(db_creds: PostgresCredentials) -> str:
        # TODO: support more database types
        """Build PostgreSQL connection string from credentials."""
        return f"postgresql://{db_creds.user}:{db_creds.password}@{db_creds.host}:{db_creds.port}/{db_creds.database}"



from typing import Any, Optional

from dbt.adapters.contracts.connection import Credentials

from ..utils import find_funcs_in_stack
from .adapter_type import ADAPTER_NAME


class DeppCredentialsWrapper:
    _db_creds: Optional[Credentials] = None

    def __init__(self, db_creds: Credentials):
        self._db_creds = db_creds

    @property
    def type(self) -> str:
        # TODO: duplicate logic also present in adapter_type.py
        if find_funcs_in_stack({"to_target_dict", "db_materialization"}):
            return self.db_creds.type
        return ADAPTER_NAME

    @property
    def db_creds(self) -> Credentials:
        if self._db_creds is None:
            raise ValueError("No valid DB Credentials")
        return self._db_creds

    def __getattr__(self, name: str) -> Any:
        """Directly proxy to the DB adapter"""
        return getattr(self._db_creds, name)



from dataclasses import dataclass
from typing import Any

from ..executors import AbstractPythonExecutor
from ..utils import get_library_from_typehint


@dataclass(frozen=True)
class ModelConfig:
    library: str
    parsed_model: dict[str, Any]
    compiled_code: str

    @classmethod
    def from_model(
        cls,
        parsed_model: dict[str, Any],
        compiled_code: str,
        default_library: str = "polars",
    ) -> "ModelConfig":
        # Resolves library from config
        library = (
            parsed_model.get("config", {}).get("library")
            or get_library_from_typehint(
                compiled_code, AbstractPythonExecutor.type_mapping
            )
            or default_library
        )
        return cls(library, parsed_model, compiled_code)



import contextlib
import inspect
from functools import lru_cache
from multiprocessing.context import SpawnContext
from pathlib import Path
from typing import Any, FrozenSet, Iterable, Type

from dbt.adapters.base.impl import BaseAdapter
from dbt.adapters.base.meta import AdapterMeta, available
from dbt.adapters.base.relation import BaseRelation
from dbt.adapters.contracts.connection import AdapterResponse, Credentials
from dbt.adapters.contracts.relation import RelationConfig
from dbt.adapters.factory import FACTORY, get_adapter_by_type  # type: ignore
from dbt.adapters.protocol import AdapterConfig
from dbt.artifacts.resources.types import ModelLanguage
from dbt.clients.jinja import MacroGenerator
from dbt.compilation import Compiler
from dbt.config.runtime import RuntimeConfig
from dbt.contracts.graph.manifest import Manifest
from dbt.parser.manifest import ManifestLoader

from .config import (
    AdapterTypeDescriptor,
    DbInfo,
    DeppCredentials,
    DeppCredentialsWrapper,
    ModelConfig,
    RelationDescriptor,
)
from .executors import *  # noqa
from .executors import AbstractPythonExecutor
from .utils.ast_utils import extract_python_docstring
from .utils.general import logs, release_plugin_lock


class DeppAdapter(metaclass=AdapterMeta):
    # TODO: fix type ignores where possible
    """DBT adapter for executing Python models with database backends."""

    Relation = RelationDescriptor()
    AdapterSpecificConfigs = AdapterConfig
    type = AdapterTypeDescriptor()
    _db_adapter_class: Type[BaseAdapter]
    _db_adapter: BaseAdapter
    db_creds: Credentials

    def __new__(cls, config: RuntimeConfig, mp_context: SpawnContext) -> "DeppAdapter":
        """Create adapter instance and configure underlying database adapter."""
        instance = super().__new__(cls)
        db_creds = cls.get_db_credentials(config)

        db_info = DbInfo.get_cached_with_relation()
        db_info.apply_overrides(config)

        with release_plugin_lock():
            db_adapter: Type[BaseAdapter] = FACTORY.get_adapter_class_by_name(  # type: ignore
                db_creds.type
            )
            original_plugin = FACTORY.get_plugin_by_name(config.credentials.type)
            original_plugin.dependencies = [db_creds.type]
            config.credentials = db_creds
            FACTORY.register_adapter(config, mp_context)
            config.credentials = DeppCredentialsWrapper(db_creds)  # type: ignore

        instance._db_adapter_class = db_adapter
        instance.db_creds = db_creds
        return instance

    def __init__(self, config: RuntimeConfig, mp_context: SpawnContext):
        """Initialize adapter with database connection and configuration."""
        self.config = config
        self.mp_context = mp_context

        # Type ignores are needed as we are overwriting some of dbt's behavior
        self._db_adapter = get_adapter_by_type(self._db_adapter_class.type())  # type: ignore
        self.connections = self._db_adapter.connections
        self._available_ = self._db_adapter._available_.union(self._available_)  # type: ignore
        self._parse_replacements_.update(self._db_adapter._parse_replacements_)  # type: ignore

    @logs
    def submit_python_job(
        self, parsed_model: dict[str, Any], compiled_code: str
    ) -> AdapterResponse:
        # TODO: Add remote executors
        """Execute Python model code selecting the requested executor."""
        config = ModelConfig.from_model(parsed_model, compiled_code)
        executor = self.get_executor(parsed_model, config.library)
        result = executor.submit(compiled_code)
        return AdapterResponse(_message=f"PYTHON | {result}")

    def get_executor(
        self, parsed_model: dict[str, Any], library: str
    ) -> AbstractPythonExecutor[Any]:
        """Get Python executor based on model's configured library"""
        registry = AbstractPythonExecutor.registry
        executor_class = registry.get(library)
        if executor_class is None:
            raise ValueError(f"No '{library}'. Available: {list(registry.keys())}")
        return executor_class(parsed_model, self.db_creds, library)

    @available
    def db_materialization(self, context: dict[str, Any], materialization: str) -> Any:
        """Execute database materialization macro."""
        macro = self.manifest.find_materialization_macro_by_name(
            self.config.project_name, materialization, self._db_adapter.type()
        )
        if macro is None:
            raise ValueError("Invalid Macro")
        return MacroGenerator(macro, context, stack=context["context_macro_stack"])()

    @classmethod
    def get_db_credentials(cls, config: RuntimeConfig) -> Credentials:
        """Extract database credentials from adapter configuration."""
        dep_credentials: DeppCredentials | DeppCredentialsWrapper = config.credentials  # type: ignore
        if isinstance(dep_credentials, DeppCredentials):
            db_info = DbInfo.get_cached_with_relation()
            return db_info.profile.credentials
        with release_plugin_lock():
            FACTORY.load_plugin(dep_credentials.db_creds.type)
        return dep_credentials.db_creds

    def get_compiler(self) -> Compiler:
        """Get DBT compiler instance for this adapter."""
        return Compiler(self.config)

    def __getattr__(self, name: str) -> Any:
        """Directly proxy to the DB adapter"""
        if hasattr(self._db_adapter, name):
            return getattr(self._db_adapter, name)
        else:
            getattr(super(), name)

    @classmethod
    def is_cancelable(cls) -> bool:
        """Python jobs cannot be cancelled once started."""
        return False

    @property
    def db_adapter(self) -> BaseAdapter:
        """Access underlying database adapter."""
        return self._db_adapter

    @property
    @lru_cache(maxsize=None)
    def manifest(self) -> Manifest:
        """Get cached DBT manifest for the project."""
        return ManifestLoader.get_full_manifest(self.config)

    def get_filtered_catalog(
        self,
        relation_configs: Iterable[RelationConfig],
        used_schemas: FrozenSet[tuple[str, str]],
        relations: set[BaseRelation] | None = None,
    ) -> tuple[Any, list[Exception]]:
        """Override to enrich Python models with docstrings"""
        for manifest in [self.manifest, self._find_parent_manifest()]:  # type: ignore
            if manifest:
                self.inject_docstring(manifest)  # type: ignore

        return self._db_adapter.get_filtered_catalog(  # type: ignore
            relation_configs, used_schemas, relations
        )

    def _find_parent_manifest(self) -> Manifest | None:
        """Find manifest from parent GenerateTask if it exists"""
        frame = inspect.currentframe()
        if frame is None:
            return None
        with contextlib.suppress(Exception):
            while frame := frame.f_back:
                if (
                    (obj := frame.f_locals.get("self"))
                    and type(obj).__name__ == "GenerateTask"
                    and getattr(obj, "manifest", None)
                ):
                    return obj.manifest  # type: ignore
        return None

    def inject_docstring(self, manifest: Manifest) -> None:
        """Extract Python model docstrings as descriptions (YAML takes precedence)"""
        project_root = Path(self.config.project_root)
        for node in manifest.nodes.values():
            if (
                getattr(node, "language", None) == ModelLanguage.python
                and node.resource_type.value == "model"
                and not (node.description or "").strip()
            ):
                if docstring := extract_python_docstring(
                    str(project_root / node.original_file_path)
                ):
                    node.description = docstring



"""Type hints for dbt Python models to improve developer experience."""

from typing import Generic, Literal, Optional, Protocol, TypeVar, Union

import geopandas as gpd
import pandas as pd
import polars as pl

from .base import DbtConfig, DbtThis

PolarsDataFrame = Union[pl.DataFrame, pl.LazyFrame]
DataFrame = Union[pd.DataFrame, PolarsDataFrame, gpd.GeoDataFrame]

DataFrameT = TypeVar("DataFrameT", bound=DataFrame, covariant=True)


class DbtObject(Protocol, Generic[DataFrameT]):
    """
    Dbt Object type providing method to reference other models and sources
    """

    def ref(
        self,
        model_name: str,
        *additional_names: str,
        version: Optional[Union[str, int]] = None,
        v: Optional[Union[str, int]] = None,
    ) -> DataFrameT:
        """Reference another model in the dbt project.

        Args:
            model_name: Name of the model to reference
            *additional_names: Additional parts of the model name (for two-part names)
            version: Model version (alternative to 'v')
            v: Model version (short form)

        Returns:
            A dataFrame containing the referenced model's data.
        """
        ...

    def source(self, source_name: str, table_name: str) -> DataFrameT:
        """Reference a source table.

        Args:
            source_name: Name of the source
            table_name: Name of the table within the source

        Returns:
            A DataFrame containing the source table's data.
        """
        ...

    def config(
        self, library: Literal["polars"] | Literal["pandas"] | Literal["geopandas"]
    ) -> DbtConfig:
        """Access model configuration."""
        ...

    @property
    def this(self) -> DbtThis:
        """Reference to the current model."""
        ...

    def is_incremental(self) -> bool:
        """Check if this is an incremental model run."""
        ...



from typing import Any, Protocol


class DbtThis:
    """Represents the current model in dbt."""

    database: str
    schema: str
    identifier: str

    def __repr__(self) -> str:
        """Return the fully qualified table name."""
        return f'"{self.database}"."{self.schema}"."{self.identifier}"'


class SessionObject(Protocol):
    """Type hints for the session object passed to Python models.

    The session object provides database connectivity for Python models.
    In most cases with this adapter, it represents a SQLAlchemy session
    or similar database connection object.
    """

    def execute(self, query: str) -> Any:
        """Execute a SQL query.

        Args:
            query: SQL query string to execute

        Returns:
            Query result
        """
        ...

    def commit(self) -> None:
        """Commit the current transaction."""
        ...

    def rollback(self) -> None:
        """Rollback the current transaction."""
        ...

    def close(self) -> None:
        """Close the session."""
        ...


class DbtConfig:
    """Configuration object for dbt models."""

    def __init__(self, *args: Any, **kwargs: Any) -> None:
        """Initialize configuration."""
        ...

    @staticmethod
    def get(key: str, default: Any | None = None) -> Any:
        """Get configuration value by key."""
        ...

    def __call__(self, **kwargs: Any) -> None:
        """Set configuration options.

        Args:
            library: DataFrame library to use ("pandas" or "polars")
            **kwargs: Additional configuration options
        """
        ...



"""Utilities for using ast to extract information out of from Python."""

import ast
import contextlib
import inspect
from typing import Optional, Union


def get_library_from_typehint(
    code: str, mapping: dict[str, str], func: str = "model"
) -> str | None:
    """Extract the library name from the model function's type annotation."""
    # TODO: find out a way to both simplify this and think about this mapping
    # TODO: Auto crete this mapping?
    for node in ast.walk(ast.parse(code)):
        if (
            isinstance(node, ast.FunctionDef)
            and node.name == func
            and node.args.args
            and (ann := node.args.args[0].annotation)
        ):
            return mapping.get(
                (
                    str(ann.value)
                    if isinstance(ann, ast.Constant)
                    else ast.unparse(ann)
                ).split(".")[-1]
            )
    return None


def get_docstring(node: Union[ast.Module, ast.FunctionDef]) -> Optional[str]:
    """Extract docstring from node if it exists."""
    if (
        node.body
        and isinstance(node.body[0], ast.Expr)
        and isinstance(node.body[0].value, ast.Constant)
        and isinstance(node.body[0].value.value, str)
    ):
        return inspect.cleandoc(node.body[0].value.value)
    return None


def extract_python_docstring(file_path: str, func_name: str = "model") -> Optional[str]:
    """Extract docstring from Python model file."""
    with contextlib.suppress(Exception):
        with open(file_path, "r", encoding="utf-8") as f:
            tree = ast.parse(f.read())

        for node in ast.walk(tree):
            if isinstance(node, ast.FunctionDef) and node.name == func_name:
                if doc := get_docstring(node):
                    return doc
        return get_docstring(tree)
    return None
